/*
 * This is the madcap library, an implementation of the MADCAP protocol for
 * multicast address allocation.
 *
 * This file contains the message handling code. For the implementation of the
 * functions defined here, please see message.c.
 *
 * Copyright (C) 2001 Hkan Bystrm,Maria Isberg, Pia Lindstm,
 * Fredrik Pettersson, Linus Svensson and ke stmark.
 *
 * The project is supported by Telia Research AB, Lule, Sweden.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 1307 USA
 *
 * The authors of this program can be reached on malloc@deimos.campus.luth.se,
 * or via the project homepage at http://deimos.campus.luth.se/malloc/.
 * The Telia representative is Thomas Eriksson (Thomas.A.Eriksson@telia.se)
 */

#ifndef _message_h_
#define _message_h_

#include "madcap.h" /* For the data types. */
#include "list.h"
#include "option.h"

/* These are the different message types used by the MADCAP protocol. */
enum{
    MC_DISCOVER = 1,
    MC_OFFER    = 2,
    MC_REQUEST  = 3,
    MC_RENEW    = 4,
    MC_ACK      = 5,
    MC_NAK      = 6,
    MC_RELEASE  = 7,
    MC_GETINFO  = 8
};

/* This is the format of a MADCAP message as represented internally. */
typedef struct mc_msg{
    uint8  version;
    uint8  msgtype;
    uint16 addrfamily;
    uint32 xid;
    list   *options;
} mc_msg;

/**
 * Used to create a REQUEST message.
 *
 * @param sid The server identifier option.
 * @param lid The lease identifier option.
 * @param scope The scope option.
 * @return The newly created message. It should be deleted using msg_delete.
 */
extern mc_msg *msg_create_request(mc_option_sid *sid,
                                  mc_option_lid *lid,
                                  mc_option_scope *scope);

/**
 * Used to create a RENEW message.
 *
 * @param lid The lease identifier option.
 * @return The newly created message. It should be deleted using msg_delete.
 */
extern mc_msg *msg_create_renew(mc_option_lid *lid);

/**
 * Used to create a RELEASE message.
 *
 * @param lid The lease identifier option.
 * @return The newly created message. It should be deleted using msg_delete.
 */
extern mc_msg *msg_create_release(mc_option_lid *lid);

/**
 * Used to create a GETINFO message.
 *
 * @param lid The lease identifier option.
 * @param reqlist The option request list option.
 * @return The newly created message. It should be deleted using msg_delete.
 */
extern mc_msg *msg_create_getinfo(mc_option_lid *lid,
                                  mc_option_reqlist *reqlist);

/**
 * This function is used to delete a message from memory. It will recursively
 * free the memory used by each option, using option_delete(), and then free
 * the message itself.
 *
 * @param message The message to clear from memory.
 */
extern void msg_delete(mc_msg *message);

/**
 * This is used to add an option to the list of options in a message.
 *
 * @param msg The message where the option should be added.
 * @param option The option which should be added.
 */
extern void msg_add_option(mc_msg *msg, mc_option *option);

/**
 * This is used to get a specific option from a message.
 *
 * @param msg The message to work with.
 * @param option The code of the option to retreive.
 * @return A pointer to the option, or NULL if the option did not exist.
 */
extern mc_option *msg_get_option(mc_msg *msg, uint16 option);

/**
 * This function is used to check the syntax of a reply message received from a
 * server.
 *
 * @param message The message to check.
 * @param original The message which resulted in the reply.
 * @return 0 if the message was ok, -1 if there was an error. Possible error
 * values are MC_EPARAM, MC_EOPTION, MC_ENOOPTION and MC_EDUPOPTION.
 */
extern int msg_validate(mc_msg *message, mc_msg *original);

#endif
