/*
 * This is the madcap library, an implementation of the MADCAP protocol for
 * multicast address allocation.
 *
 * This file contains the implementation of the network handling code.
 *
 * Copyright (C) 2001 Hkan Bystrm,Maria Isberg, Pia Lindstm,
 * Fredrik Pettersson, Linus Svensson and ke stmark.
 *
 * The project is supported by Telia Research AB, Lule, Sweden.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 1307 USA
 *
 * The authors of this program can be reached on malloc@deimos.campus.luth.se,
 * or via the project homepage at http://deimos.campus.luth.se/malloc/.
 * The Telia representative is Thomas Eriksson (Thomas.A.Eriksson@telia.se)
 */

#include "net.h"
#include "madcap.h" /* For the data types. */
#include "message.h"
#include "debug.h"
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <unistd.h> /* For alarm(). */
#include <signal.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>

/* Add 1, 2 or 4 bytes to a raw message. hton will be used for 2 and 4. */
static uint8 *add_raw8(uint8 data, uint8 *raw_message, int *index);
static uint8 *add_raw16(uint16 data, uint8 *raw_message, int *index);
static uint8 *add_raw32(uint32 data, uint8 *raw_message, int *index);

/* Add a block of bytes to a raw message. hton is _not_ used. */
static uint8 *add_raw(uint8 *data, uint8 *raw_message, int *index, unsigned int count);

/* Read 1, 2 or 4 bytes from a raw message. ntoh will be used for 2 and 4. */
static uint8  read_raw8(uint8 *raw_message, int *index);
static uint16 read_raw16(uint8 *raw_message, int *index);
static uint32 read_raw32(uint8 *raw_message, int *index);

/* Read a block of bytes from a raw message. ntoh is _not_ used. */
static uint8 *read_raw(uint8 *raw_message, int *index, unsigned int count);

/**
 * This function is used to encode a message. The message will be copied to a
 * block of memory and will there have the same representation as it should have
 * when sent over the network.
 *
 * @param message The message to encode.
 * @param message_size The size if the encoded message will be stored here.
 * @return A block of memory containing the message in network format.
 */
uint8 *mc_encode_message(mc_msg *message, int *message_size){
    uint8 *raw_message = NULL;
    int index = 0;
    mc_option *option = NULL;

    /* Memory for the basic fields. */
    raw_message = (uint8 *)malloc(8 * sizeof(uint8));
    ASSERT(raw_message != NULL, "Memory allocation failed.");

    *(raw_message + index) = (uint8)MC_VERSION;
    index++;
    *(raw_message + index) = message->msgtype;
    index++;
    *(uint16 *)(raw_message + index) = htons(message->addrfamily);
    index = index + 2;
    *(uint32 *)(raw_message + index) = htonl(message->xid);
    index = index + 4;

    option = list_first(message->options);
    while(option != NULL) {
        raw_message = mc_encode_option(option, raw_message, &index);
        option = list_next(message->options);
    }

    *message_size = index;
    return raw_message;
}

/**
 * This function is used to encode an option and add it to the end of an
 * existing, already encoded, message.
 *
 * @param option The option to encode.
 * @param raw_message The message to add to option to.
 * @param index Where in the message to add the option. The index will be
 * updated as a result of calling this function.
 * @return The newly encoded message. The pointer may be the same as the
 * parameter message, but may also have changed.
 */
uint8 *mc_encode_option(mc_option *option, uint8 *raw_message, int *index){
    switch(option->code) {
    case MC_END:
        raw_message = add_raw16(((mc_option_end *)option)->code,
                                raw_message, index);
        raw_message = add_raw16(((mc_option_end *)option)->len,
                                raw_message, index);
        break;
    case MC_LTIME:
        raw_message = add_raw16(((mc_option_ltime *)option)->code,
                                raw_message, index);
        raw_message = add_raw16(((mc_option_ltime *)option)->len,
                                raw_message, index);
        raw_message = add_raw32(((mc_option_ltime *)option)->lease_time,
                                raw_message, index);
        break;
    case MC_SID:
        raw_message = add_raw16(((mc_option_sid *)option)->code,
                                raw_message, index);
        raw_message = add_raw16(((mc_option_sid *)option)->len,
                                raw_message, index);
        raw_message = add_raw16(((mc_option_sid *)option)->addr_family,
                                raw_message, index);
        raw_message = add_raw32(((mc_option_sid *)option)->address,
                                raw_message, index);
        break;
    case MC_LID:
        raw_message = add_raw16(((mc_option_lid *)option)->code,
                                raw_message, index);
        raw_message = add_raw16(((mc_option_lid *)option)->len,
                                raw_message, index);
        raw_message = add_raw(((mc_option_lid *)option)->id,
                              raw_message, index,
                              ((mc_option_lid *)option)->len);
        break;
    case MC_SCOPE:
        raw_message = add_raw16(((mc_option_scope *)option)->code,
                                raw_message, index);
        raw_message = add_raw16(((mc_option_scope *)option)->len,
                                raw_message, index);
        raw_message = add_raw32(((mc_option_scope *)option)->scope_id,
                                raw_message, index);
        break;
    case MC_REQLIST:
        raw_message = add_raw16(((mc_option_reqlist *)option)->code,
                                raw_message, index);
        raw_message = add_raw16(((mc_option_reqlist *)option)->len,
                                raw_message, index);

        {
            uint16 *opt =
                (uint16 *)list_first(((mc_option_reqlist *)option)->options);
            while(opt != NULL) {
                raw_message = add_raw16(*opt, raw_message, index);
                opt =
                    (uint16 *)list_next(((mc_option_reqlist *)option)->options);
            }
        }
        break;
    case MC_STIME:
        raw_message = add_raw16(((mc_option_stime *)option)->code,
                                raw_message, index);
        raw_message = add_raw16(((mc_option_stime *)option)->len,
                                raw_message, index);
        raw_message = add_raw32(((mc_option_stime *)option)->start_time,
                                raw_message, index);
        break;
    case MC_NUMADDR:
        raw_message = add_raw16(((mc_option_numaddr *)option)->code,
                                raw_message, index);
        raw_message = add_raw16(((mc_option_numaddr *)option)->len,
                                raw_message, index);
        raw_message = add_raw16(((mc_option_numaddr *)option)->min,
                                raw_message, index);
        raw_message = add_raw16(((mc_option_numaddr *)option)->desired,
                                raw_message, index);
        break;
    case MC_REQLANG:
        raw_message = add_raw16(((mc_option_reqlang *)option)->code,
                                raw_message, index);
        raw_message = add_raw16(((mc_option_reqlang *)option)->len,
                                raw_message, index);
        raw_message = add_raw(((mc_option_reqlang *)option)->tag,
                              raw_message, index,
                              ((mc_option_reqlang *)option)->len);
        break;
    case MC_SCOPELIST:
        ABORT("Attempt to encode Multicast Scope List option.");
        break;
    case MC_ADDRLIST:
        raw_message = add_raw16(((mc_option_addrlist *)option)->code,
                                raw_message, index);
        raw_message = add_raw16(((mc_option_addrlist *)option)->len,
                                raw_message, index);
        {
            address_range *range =
                (address_range *)list_first(((mc_option_addrlist *)option)->
                                            ranges);
            while(range != NULL) {
                raw_message = add_raw32(range->start, raw_message, index);
                raw_message = add_raw16((uint16)(range->end - range->start + 1),
                                        raw_message, index);
                range = list_next(((mc_option_addrlist *)option)->ranges);
            }
        }
        break;
    case MC_CTIME:
        raw_message = add_raw16(((mc_option_ctime *)option)->code,
                                raw_message, index);
        raw_message = add_raw16(((mc_option_ctime *)option)->len,
                                raw_message, index);
        raw_message = add_raw32(((mc_option_ctime *)option)->time,
                                raw_message, index);
        break;
    case MC_FEATURELIST:
        raw_message = add_raw16(((mc_option_featurelist *)option)->code,
                                raw_message, index);
        raw_message = add_raw16(
            list_count(((mc_option_featurelist *)option)->supported) * 2 +
            list_count(((mc_option_featurelist *)option)->requested) * 2 +
            list_count(((mc_option_featurelist *)option)->required) * 2 + 6,
            raw_message, index);
        {
            int tmp_index = 0;
            uint16 tmp_nsupported =
                list_count(((mc_option_featurelist *)option)->supported);
            uint16 tmp_nrequested =
                list_count(((mc_option_featurelist *)option)->requested);
            uint16 tmp_nrequired =
                list_count(((mc_option_featurelist *)option)->required);
            raw_message = add_raw16(tmp_nsupported, raw_message, index);
            while(tmp_index < tmp_nsupported) {
                raw_message = add_raw16(
                    *(uint16 *)list_jump(
                        ((mc_option_featurelist *)option)->supported,
                        tmp_index),
                    raw_message, index);
                tmp_index++;
            }
            index = 0;
            raw_message = add_raw16(tmp_nrequested, raw_message, index);
            while(tmp_index < tmp_nrequested) {
                raw_message = add_raw16(
                    *(uint16 *)list_jump(
                        ((mc_option_featurelist *)option)->requested,
                        tmp_index),
                    raw_message, index);
                tmp_index++;
            }
            index = 0;
            raw_message = add_raw16(tmp_nrequired, raw_message, index);
            while(tmp_index < tmp_nrequired){
                raw_message = add_raw16(
                    *(uint16 *)list_jump(
                        ((mc_option_featurelist *)option)->required, tmp_index),
                    raw_message, index);
                tmp_index++;
            }
        }
        break;
    case MC_RTIME:
        ABORT("Attempt to encode Retry Time option.");
        break;
    case MC_MIN_LTIME:
        raw_message = add_raw16(((mc_option_min_ltime *)option)->code,
                                raw_message, index);
        raw_message = add_raw16(((mc_option_min_ltime *)option)->len,
                                raw_message, index);
        raw_message = add_raw32(((mc_option_min_ltime *)option)->min_ltime,
                                raw_message, index);
        break;
    case MC_MAX_STIME:
        raw_message = add_raw16(((mc_option_max_stime *)option)->code,
                                raw_message, index);
        raw_message = add_raw16(((mc_option_max_stime *)option)->len,
                                raw_message, index);
        raw_message = add_raw32(((mc_option_max_stime *)option)->max_stime,
                                raw_message, index);
        break;
    default:
        ABORT("Unknown option type or encoding not enabled for option.");
        break;
    }
    return raw_message;
}

/**
 * This function will send an encoded message to a MAAS server or to a multicast
 * address.
 *
 * @param data A pointer to the encoded data.
 * @param size The number of bytes in the message.
 * @param address The address to the server or multicast group. It should be in
 * host byte order.
 * @param port The port the message will be sent over. It should be in host byte
 * order.
 * @param ttl The time-to-live which should be used for this message if it is
 * sent over a multicast address.
 * @return The socket where the message was sent, or -1 if there was a problem.
 * Possible errors are MC_EPARAM, MC_EFAIL and MC_ESIZE.
 */
int mc_send_message(uint8 *data, int size,
                    uint32 address, uint16 port, uint8 ttl){
    int sock = -1; /* Local socket descriptor. */
    struct sockaddr_in sa; /* Server address here. */
    struct sockaddr_in ca; /* Client address here. */
    memset(&sa, 0, sizeof(sa));
    memset(&ca, 0, sizeof(ca));
    
    if(size <= 0) {
        DEBUG("size <= 0 in mc_send_message.");
        mc_errno = MC_EPARAM;
        return -1;
    }
    if(data == NULL) {
        DEBUG("data == NULL in mc_send_message.");
        mc_errno = MC_EPARAM;
        return -1;
    }

    sock = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if(sock == -1) {
        DEBUG("Creation of socket failed in mc_send_message.");
        mc_errno = MC_EFAIL;
        return -1;
    }

    ca.sin_family = PF_INET;
    ca.sin_addr.s_addr = INADDR_ANY;
    ca.sin_port = INADDR_ANY;

    if(bind(sock, (struct sockaddr *)&ca, sizeof(ca)) == -1) {
        DEBUG("Binding of socket failed in mc_send_message.");
        mc_errno = MC_EFAIL;
        return -1;
    }

    sa.sin_family = PF_INET;
    sa.sin_addr.s_addr = htonl(address);
    sa.sin_port = htons(port);

    if(ttl != 0) {
        /* Set the ttl of the socket. */
#ifdef MC_FreeBSD
        if(setsockopt(sock, IPPROTO_IP, IP_MULTICAST_TTL, &ttl,
                      sizeof(ttl)) == -1) {
#else
        if(setsockopt(sock, SOL_IP, IP_MULTICAST_TTL, &ttl,
                      sizeof(ttl)) == -1) {
#endif
            DEBUG("Failed to set the ttl of sock in mc_send_message.");
            mc_errno = MC_EFAIL;
            return -1;
        }
    }

    if(sendto(sock, data, size, 0, (struct sockaddr *)&sa, sizeof(sa)) < size) {
        DEBUG("Too little data sent in mc_send_message.");
        mc_errno = MC_ESIZE;
        return -1;
    }
    
    return sock;
}

/* Handle SIGALRM for receive time-out. */
void catch_alrm(int sig){
    DEBUG("SIGALRM received for mc_receive_message.");
}

/**
 * This function will receive a message from the network.
 *
 * @param socket The socket where the message should be received. The socket
 * will not be closed after a successful reception.
 * @param timeout The period of time to wait before failing with an error. Time
 * is in seconds.
 * @param message_size The size of the received message will be stored here.
 * @return A pointer to the raw data received from the network. NULL if the
 * reception failed. Possible errors are MC_EPARAM, MC_EINTR, MC_EFAIL,
 * MC_ETIMEOUT and MC_ESIZE.
 */
uint8 *mc_receive_message(int socket, unsigned int timeout, int *message_size){
    struct sockaddr_in sa; /* Server address here. */
#ifdef MC_FreeBSD
    int fromsize = sizeof(sa);
#else
    socklen_t fromsize = sizeof(sa);
#endif
    int size = 0;

    static struct sigaction act, oldact;

    uint8 *buffer = NULL;
    
    if(socket == -1) {
        DEBUG("Invalid socket in mc_receive_message.");
        mc_errno = MC_EPARAM;
        return NULL;
    }

    buffer = (uint8 *)malloc(MC_RECVBUFSIZE * sizeof(uint8));
    ASSERT(buffer != NULL, "Memory allocation failed.");

    /* It would be great if we could do this without alarm, since it is not ansi
       C. Poll may be an option, but I think that is non-ansi as well. */
    /* Handle SIGALRM: */
    act.sa_handler = catch_alrm;
    sigaction(SIGALRM, &act, &oldact);

    /* Set alarm for the timeout. */
    alarm(timeout);

    size = recvfrom(socket, buffer, MC_RECVBUFSIZE, 0,
                    (struct sockaddr *)&sa, &fromsize);

    /* Turn off alarm. */
    timeout = alarm(0);
    /* Restore old way of handling SIGALRM. */
    sigaction(SIGALRM, &oldact, NULL);
    
    if(size == -1) {
        if(errno == EINTR) {
            /* We were interrupted by a signal. */
            if(timeout == 0) {
                DEBUG("Timeout in mc_receive_message");
                free(buffer);
                mc_errno = MC_ETIMEOUT;
                return NULL;
            } else {
                DEBUG("Unknown interruption in mc_receive_message");
                free(buffer);
                mc_errno = MC_EINTR;
                return NULL;
            }
        } else {
            DEBUG("Failed to receive data in mc_receive_message");
            free(buffer);
            mc_errno = MC_EFAIL;
            return NULL;
        }
    } else if(size == MC_RECVBUFSIZE) {
        DEBUG("Buffer full in mc_receive_message");
        free(buffer);
        mc_errno = MC_ESIZE;
        return NULL;
    }

    *message_size = size;
    return buffer;
}

/**
 * This function will decode a message as it comes from the receive function and
 * create a mc_msg to store it in. No validation is done here.
 *
 * @param raw_message The raw message.
 * @param size The number of bytes in the message.
 * @return A pointer to the decoded message, or NULL if decoding failed.
 * Possible errors are MC_EBROKEN and MC_EVERSION.
 */
mc_msg *mc_decode_message(uint8 *raw_message, int size){
    int index = 0;
    mc_msg *message = (mc_msg *)malloc(sizeof(mc_msg));
    ASSERT(message != NULL, "Memory allocation failed.");

    ASSERT(raw_message != NULL, "raw_message == NULL in mc_decode_message.");

    ASSERT(size > 0, "size <= 0 in mc_decode_message.");

    message->version = read_raw8(raw_message, &index);
    if(message->version != MC_VERSION) {
        DEBUG("Wrong version of MADCAP in mc_decode_message.");
        msg_delete(message);
        mc_errno = MC_EVERSION;
        return NULL;
    }

    message->msgtype = read_raw8(raw_message, &index);
    message->addrfamily = read_raw16(raw_message, &index);
    message->xid = read_raw32(raw_message, &index);

    message->options = list_create();
    ASSERT(message->options != NULL, "List creation failed.");

    /* An option is at least 4 bytes, code + len. */
    while((size - index) >= 4) {
        mc_option *opt = mc_decode_option(raw_message, size, &index);
        if(opt == NULL) {
            DEBUG("Failed tp decode option in mc_decode_message.");
            msg_delete(message);
            mc_errno = MC_EBROKEN;
            return NULL;
        }
        list_insert(message->options, opt, LIST_AFTER);
    }

    if(size != index) {
        DEBUG("There was extra data after last option in mc_decode_message.");
        msg_delete(message);
        mc_errno = MC_EBROKEN;
        return NULL;
    }
    
    return message;
}

/**
 * This function will decode an option as it comes from the receive function and
 * create a mc_option to store it in. It is used by mc_decode_message.
 *
 * @param raw_message The raw message.
 * @param size The number of bytes in the message.
 * @return A pointer to the decoded message, or NULL if decoding failed.
 */
mc_option *mc_decode_option(uint8 *raw_message, int size, int *index){
    uint16 code;
    uint16 len;
    mc_option *opt = NULL;
    ASSERT(raw_message != NULL, "raw_message == NULL in mc_decode_option");
    ASSERT(size > 0, "size <= 0 in mc_decode_option");
    ASSERT(index != NULL, "index == NULL in mc_decode_option");
    ASSERT((size - *index) >= 4,
           "Too little space for option in mc_decode_option.");

    code = read_raw16(raw_message, index);
    len = read_raw16(raw_message, index);

    if((size - *index) < len) {
        DEBUG("Small size of data of option in mc_decode_option.");
        return NULL;
    }
    
    switch(code) {
    case MC_END:
        if(len != 0) {
            DEBUG("len != 0 of MC_END option in mc_decode_option.");
            return NULL;
        }
        opt = (mc_option *)option_create_end();
        DEBUG("Decoded End option.");
        break;
    case MC_LTIME:
        if(len != 4) {
            DEBUG("len != 4 of MC_LTIME option in mc_decode_option.");
            return NULL;
        }
        opt = (mc_option *)option_create_ltime(read_raw32(raw_message, index));
        DEBUG("Decoded Lease Time option.");
        break;
    case MC_SID:
        if(len != 6) {
            DEBUG("len != 6 of MC_SID option in mc_decode_option.");
            return NULL;
        }
        if(read_raw16(raw_message, index) != MC_IPV4) {
            DEBUG("Family not IPv4 of MC_SID option in mc_decode_option.");
            return NULL;
        }
        opt = (mc_option *)option_create_sid(read_raw32(raw_message, index));
        DEBUG("Decoded Server ID option.");
        break;
    case MC_LID:
        if(len < 1) {
            DEBUG("len < 1 of MC_LID option in mc_decode_option.");
            return NULL;
        }
        opt = (mc_option *)
            option_create_lid(read_raw(raw_message, index, len), len);
        DEBUG("Decoded Lease ID option.");
        break;
    case MC_SCOPE:
        if(len != 4) {
            DEBUG("len != 4 of MC_SCOPE option in mc_decode_option.");
            return NULL;
        }
        opt = (mc_option *)option_create_scope(read_raw32(raw_message, index));
        DEBUG("Decoded Scope option.");
        break;
    case MC_REQLIST:
        if(len < 2) {
            DEBUG("len < 2 of MC_REQLIST option in mc_decode_option.");
            return NULL;
        }
        opt = (mc_option *)
            option_create_reqlist(len,
                                  (uint16 *)read_raw(raw_message, index, len));
        DEBUG("Decoded Option Request List option.");
        break;
    case MC_STIME:
        if(len != 4) {
            DEBUG("len != 4 of MC_STIME option in mc_decode_option.");
            return NULL;
        }
        opt = (mc_option *)option_create_stime(read_raw32(raw_message, index));
        DEBUG("Decoded Start Time option.");
        break;
    case MC_NUMADDR:
        if(len != 4) {
            DEBUG("len != 4 of MC_NUMADDR option in mc_decode_option.");
            return NULL;
        }
        opt = (mc_option *)option_create_numaddr(read_raw16(raw_message,
                                                            index),
                                                 read_raw16(raw_message,
                                                            index));
        DEBUG("Decoded Number of Addresses option.");
        break;
    case MC_REQLANG:
        opt = (mc_option *)option_create_reqlang(len, read_raw(raw_message,
                                                               index, len));
        DEBUG("Decoded Requested Language option.");
        break;
    case MC_SCOPELIST:
    {
        list *tmp_scopes = list_create();
        uint8 tmp_i = read_raw8(raw_message, index);
        uint8 tmp_j = 0;
        while(tmp_i > 0){
            scope *tmp_scope = (scope *)malloc(sizeof(scope));
            tmp_scope->addresses.start = read_raw32(raw_message, index);
            tmp_scope->addresses.end = read_raw32(raw_message, index);
            tmp_scope->ttl = read_raw8(raw_message, index);
            tmp_scope->num_names = read_raw8(raw_message, index);
            tmp_scope->name_set =
                (name *)malloc(tmp_scope->num_names * sizeof(name));
            while(tmp_j < tmp_scope->num_names){
                /* TODO: We ignore the flags field at the moment. */
                uint8 lang_len = 0, name_len = 0;
                read_raw8(raw_message, index);
                lang_len = read_raw8(raw_message, index);
                tmp_scope->name_set[tmp_j].language_tag =
                    (char *)malloc((lang_len + 1) * sizeof(char));
                memcpy(tmp_scope->name_set[tmp_j].language_tag,
                       read_raw(raw_message, index, lang_len),
                       lang_len * sizeof(char));
                tmp_scope->name_set[tmp_j].language_tag[lang_len] = '\0';
                name_len = read_raw8(raw_message, index);
                tmp_scope->name_set[tmp_j].name =
                    (char *)malloc((name_len + 1) * sizeof(char));
                memcpy(tmp_scope->name_set[tmp_j].name,
                       read_raw(raw_message, index, name_len),
                       name_len * sizeof(char));
                tmp_scope->name_set[tmp_j].name[name_len] = '\0';
                tmp_j++;
            }
            list_insert(tmp_scopes, tmp_scope, LIST_AFTER);
            tmp_i--;
        }
        opt = (mc_option *)option_create_scopelist(len, tmp_scopes);
        break;
    }
    case MC_ADDRLIST:
    {
        int tmp_len = len;
        list *tmp_ranges = list_create();
        ASSERT(tmp_ranges != NULL, "Failed to create list.");
        while(tmp_len > 0) {
            address_range *tmp_range =
                (address_range *)malloc(sizeof(address_range));;
            ASSERT(tmp_range != NULL, "Memory allocation failed.");
            tmp_range->start = read_raw32(raw_message, index);
            tmp_range->end = tmp_range->start +
                read_raw16(raw_message, index) - 1;
            tmp_len = tmp_len - 6;
            list_insert(tmp_ranges, tmp_range, LIST_AFTER);
        }
        opt = (mc_option *)option_create_addrlist(len / 6, tmp_ranges);
        DEBUG("Decoded Address List option.");
        break;
    }
    case MC_CTIME:
        if(len != 4) {
            DEBUG("len != 4 of MC_CTIME option in mc_decode_option.");
            return NULL;
        }
        opt = (mc_option *)option_create_ctime(read_raw32(raw_message, index));
        DEBUG("Decoded Current Time option.");
        break;
    case MC_FEATURELIST:
        if(len < 6) {
            DEBUG("len < 6 of MC_FEATURELIST option in mc_decode_option.");
            return NULL;
        }
        {
            uint16 tmp_numfeatures = read_raw16(raw_message, index);
            uint16 tmp_len = len - 2;
            list *tmp_supported = list_create();
            list *tmp_requested = list_create();
            list *tmp_required = list_create();
            ASSERT((tmp_supported != NULL) && (tmp_requested != NULL) &&
                   (tmp_required != NULL), "Failed to create list.");

            if((4 + tmp_numfeatures * 2) > tmp_len) {
                DEBUG("Number of supported features too large.");
                return NULL;
            }
            tmp_len = tmp_len - tmp_numfeatures * 2;

            while(tmp_numfeatures > 0) {
                uint16 *tmp_feature = (uint16 *)malloc(sizeof(uint16));
                ASSERT(tmp_feature != NULL, "Failed to allocate memory.");
                list_insert(tmp_supported, tmp_feature, LIST_AFTER);
                tmp_numfeatures--;
            }
            
            tmp_numfeatures = read_raw16(raw_message, index);
            tmp_len = tmp_len - 2;
            if((4 + tmp_numfeatures * 2) > tmp_len) {
                DEBUG("Number of requested features too large.");
                return NULL;
            }
            tmp_len = tmp_len - tmp_numfeatures * 2;

            while(tmp_numfeatures > 0) {
                uint16 *tmp_feature = (uint16 *)malloc(sizeof(uint16));
                ASSERT(tmp_feature != NULL, "Failed to allocate memory.");
                list_insert(tmp_requested, tmp_feature, LIST_AFTER);
                tmp_numfeatures--;
            }

            tmp_numfeatures = read_raw16(raw_message, index);
            tmp_len = tmp_len - 2;
            if((4 + tmp_numfeatures * 2) > tmp_len) {
                DEBUG("Number of required features too large.");
                return NULL;
            }
            tmp_len = tmp_len - tmp_numfeatures * 2;
            if(tmp_len > 0) {
                DEBUG("Space left after required features.");
                return NULL;
            }

            while(tmp_numfeatures > 0) {
                uint16 *tmp_feature = (uint16 *)malloc(sizeof(uint16));
                ASSERT(tmp_feature != NULL, "Failed to allocate memory.");
                list_insert(tmp_required, tmp_feature, LIST_AFTER);
                tmp_numfeatures--;
            }

            opt = (mc_option *)option_create_featurelist(tmp_supported,
                                                         tmp_requested,
                                                         tmp_required);
        }
        break;
    case MC_RTIME:
        if(len != 4) {
            DEBUG("len != 4 of MC_RTIME option in mc_decode_option.");
            return NULL;
        }
        opt = (mc_option *)option_create_rtime(read_raw32(raw_message, index));
        DEBUG("Decoded Retry Time option.");
        break;
    case MC_MIN_LTIME:
        if(len != 4) {
            DEBUG("len != 4 of MC_MIN_LTIME option in mc_decode_option.");
            return NULL;
        }
        opt = (mc_option *)option_create_min_ltime(read_raw32(raw_message,
                                                              index));
        DEBUG("Decoded Min Lease Time option.");
        break;
    case MC_MAX_STIME:
        if(len != 4) {
            DEBUG("len != 4 of MC_MAX_STIME option in mc_decode_option.");
            return NULL;
        }
        opt = (mc_option *)option_create_max_stime(read_raw32(raw_message,
                                                              index));
        DEBUG("Decoded Max Start Time option.");
        break;
    case MC_ERROR:
        if(len < 2) {
            DEBUG("len < 2 of MC_ERROR option in mc_decode_option.");
            return NULL;
        }
        {
            uint16 tmp_ecode = read_raw16(raw_message, index);
            uint8* tmp_data = read_raw(raw_message, index, len - 2);
            opt = (mc_option *)option_create_error(tmp_ecode, tmp_data,
                                                   len - 2);
        }
        DEBUG("Decoded Error option.");
        break;
    default:
        DEBUG("Unknown option code in mc_decode_option.");
        opt =  NULL;
    }
    
    return opt;
}

/* To remember the current xid. */
static uint32 xid = 0;

/**
 * Get a new transaction identifier.
 *
 * @return The new transaction identifier.
 */
uint32 mc_get_xid(){
    /* If xid is 0 we assume we have just started. May be wrong, but it does not
       matter. */
    if(xid == 0) {
        srand(time(NULL));
        *((uint8 *)&xid) = (uint8)rand();
        *((uint8 *)(&xid) + 1) = (uint8)rand();
        *((uint8 *)(&xid) + 2) = (uint8)rand();
        *((uint8 *)(&xid) + 3) = (uint8)rand();
    } else {
        xid++;
    }
    return xid;
}

static uint8 *add_raw8(uint8 data, uint8 *raw_message, int *index){
    raw_message = (uint8 *)realloc(raw_message,
                                   (*index + 1) * sizeof(uint8));
    ASSERT(raw_message != NULL, "Memory reallocation failed.");
    *(uint8 *)(raw_message + *index) = data;
    *index = *index + 1;
    return raw_message;
}

static uint8 *add_raw16(uint16 data, uint8 *raw_message, int *index){
    raw_message = (uint8 *)realloc(raw_message,
                                   (*index + 2) * sizeof(uint8));
    ASSERT(raw_message != NULL, "Memory reallocation failed.");
    *(uint16 *)(raw_message + *index) = htons(data);
    *index = *index + 2;
    return raw_message;
}

static uint8 *add_raw32(uint32 data, uint8 *raw_message, int *index){
    raw_message = (uint8 *)realloc(raw_message,
                                   (*index + 4) * sizeof(uint8));
    ASSERT(raw_message != NULL, "Memory reallocation failed.");
    *(uint32 *)(raw_message + *index) = htonl(data);
    *index = *index + 4;
    return raw_message;
}

static uint8 *add_raw(uint8 *data, uint8 *raw_message, int *index, unsigned int count){
    raw_message = (uint8 *)realloc(raw_message,
                                   (*index + count) * sizeof(uint8));
    ASSERT(raw_message != NULL, "Memory reallocation failed.");
    memcpy(raw_message + *index, data, count);
    *index = *index + count;
    return raw_message;
}

static uint8 read_raw8(uint8 *raw_message, int *index){
    uint8 ret = *(raw_message + *index);
    *index = *index + 1;
    return ret;
}

static uint16 read_raw16(uint8 *raw_message, int *index){
    uint16 ret = *(uint16 *)(raw_message + *index);
    *index = *index + 2;
    return ntohs(ret);
}

static uint32 read_raw32(uint8 *raw_message, int *index){
    uint32 ret = *(uint32 *)(raw_message + *index);
    *index = *index + 4;
    return ntohl(ret);
}

static uint8 *read_raw(uint8 *raw_message, int *index, unsigned int count){
    uint8 *ret = (uint8 *)malloc(count * sizeof(uint8));
    ASSERT(ret != NULL, "Memory allocation failed.");
    memcpy(ret, raw_message + *index, count);
    *index = *index + count;
    return ret;
}
