/*
 * This is the madcap library, an implementation of the MADCAP protocol for
 * multicast address allocation.
 *
 * This file contains the implementation of the option handling code.
 *
 * Copyright (C) 2001 Hkan Bystrm,Maria Isberg, Pia Lindstm,
 * Fredrik Pettersson, Linus Svensson and ke stmark.
 *
 * The project is supported by Telia Research AB, Lule, Sweden.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 1307 USA
 *
 * The authors of this program can be reached on malloc@deimos.campus.luth.se,
 * or via the project homepage at http://deimos.campus.luth.se/malloc/.
 * The Telia representative is Thomas Eriksson (Thomas.A.Eriksson@telia.se)
 */

#include "option.h"
#include "madcap.h" /* For the data types. */
#include "list.h"
#include "debug.h"
#include <stdlib.h>
#include <string.h>
#include <time.h>

/**
 * Used to create an End option.
 *
 * @return An End option.
 */
mc_option_end *option_create_end(){
    mc_option_end *end_opt = (mc_option_end *)malloc(sizeof(mc_option_end));
    ASSERT(end_opt != NULL, "Memory allocation failed.");

    end_opt->code       = MC_END;
    end_opt->len        = 0;

    return end_opt;
}

/**
 * Used to create a Lease Time option.
 *
 * @param lease_time The lease time in units of seconds.
 * @return A Lease Time option.
 */
mc_option_ltime *option_create_ltime(uint32 lease_time){
    mc_option_ltime *ltime_opt =
        (mc_option_ltime *)malloc(sizeof(mc_option_ltime));
    ASSERT(ltime_opt != NULL, "Memory allocation failed.");

    ltime_opt->code       = MC_LTIME;
    ltime_opt->len        = 4;
    ltime_opt->lease_time = lease_time;

    return ltime_opt;
}

/**
 * Used to create a Server ID option.
 *
 * @param address The address of the server.
 * @return A Server ID option.
 */
mc_option_sid *option_create_sid(uint32 address){
    mc_option_sid *sid_opt = (mc_option_sid *)malloc(sizeof(mc_option_sid));
    ASSERT(sid_opt != NULL, "Memory allocation failed.");
    sid_opt->code        = MC_SID;
    sid_opt->len         = 6;
    sid_opt->addr_family = MC_IPV4;
    sid_opt->address     = address;

    return sid_opt;
}

/* Keep track of if the rand function is inited. */
int lid_randomizer_initialized = 0;

/**
 * Used to create a Lease ID option.
 *
 * @param id The id to use in the option. If it is NULL then a 16 byte id will
 * be created.
 * @param len The length of the id option.
 * @return A Lease ID option.
 */
mc_option_lid *option_create_lid(uint8 *id, unsigned int len){
    mc_option_lid *lid_opt = NULL;
    
    lid_opt = (mc_option_lid *)malloc(sizeof(mc_option_lid));
    ASSERT(lid_opt != NULL, "Memory allocation failed.");
    lid_opt->code = MC_LID;

    if(id != NULL) {
        lid_opt->len = len;
        lid_opt->id = (uint8 *)malloc(len * sizeof(uint8));
        ASSERT(lid_opt->id != NULL, "Memory allocation failed.");
        memcpy(lid_opt->id, id, len);
    } else {
        len = 16;
        lid_opt->len  = len;
        lid_opt->id   = (uint8 *)malloc(len * sizeof(uint8));
        ASSERT(lid_opt->id != NULL, "Memory allocation failed.");

        if(!lid_randomizer_initialized) {
            srand(time(NULL));
            lid_randomizer_initialized = 1;
        }
        for(;len > 0;len--) {
            *(lid_opt->id + len - 1) = (uint8)rand();
        }
    }

    return lid_opt;
}

/**
 * Used to create a Scope option.
 *
 * @param scope The first address in the scope.
 * @return A Scope option.
 */
mc_option_scope *option_create_scope(uint32 scope){
    mc_option_scope *scope_opt =
        (mc_option_scope *)malloc(sizeof(mc_option_scope));
    ASSERT(scope_opt != NULL, "Memory allocation failed.");

    scope_opt->code     = MC_SCOPE;
    scope_opt->len      = 4;
    scope_opt->scope_id = scope;

    return scope_opt;
}

/**
 * Used to create an Option Request List option.
 *
 * @param len The number of options in the list.
 * @param options The list of requested options.
 * @return An Option Request List option.
 */
mc_option_reqlist *option_create_reqlist(uint16 len, uint16 *options){
    uint16 index = 0;
    mc_option_reqlist *reqlist_opt =
        (mc_option_reqlist *)malloc(sizeof(mc_option_reqlist));
    ASSERT(reqlist_opt != NULL, "Memory allocation failed.");

    reqlist_opt->code     = MC_REQLIST;
    reqlist_opt->len      = len * 2;
    reqlist_opt->options = list_create();
    while(index < len) {
        uint16 *tmp_opt = (uint16 *)malloc(sizeof(uint16));
        ASSERT(tmp_opt != NULL, "Memory allocation failed.");
        *tmp_opt = options[index];
        list_insert(reqlist_opt->options, tmp_opt, LIST_AFTER);
        index++;
    }
    
    return reqlist_opt;
}

/**
 * Used to create a Start Time option.
 *
 * @param start_time The start time.
 * @return A Start Time option.
 */
mc_option_stime *option_create_stime(uint32 start_time){
    mc_option_stime *stime_opt =
        (mc_option_stime *)malloc(sizeof(mc_option_stime));
    ASSERT(stime_opt != NULL, "Memory allocation failed.");

    stime_opt->code       = MC_STIME;
    stime_opt->len        = 4;
    stime_opt->start_time = start_time;

    return stime_opt;
}

/**
 * Used to create a Number of Addresses Requested option.
 *
 * @param min The minimum number of addresses.
 * @param desired The desired number of addresses.
 * @return A Number of Addresses Requested option.
 */
mc_option_numaddr *option_create_numaddr(uint16 min, uint16 desired){
    mc_option_numaddr *numaddr_opt =
        (mc_option_numaddr *)malloc(sizeof(mc_option_numaddr));
    ASSERT(numaddr_opt != NULL, "Memory allocation failed.");

    numaddr_opt->code    = MC_NUMADDR;
    numaddr_opt->len     = 4;
    numaddr_opt->min     = min;
    numaddr_opt->desired = desired;

    return numaddr_opt;
}

/**
 * Used to create a Requested Language option.
 *
 * @param len The number of language tags.
 * @param tag The list of language tags.
 * @return A Requested Language option.
 */
mc_option_reqlang *option_create_reqlang(uint16 len, uint8 *tag){
    mc_option_reqlang *reqlang_opt =
        (mc_option_reqlang *)malloc(sizeof(mc_option_reqlang));
    ASSERT(reqlang_opt != NULL, "Memory allocation failed.");
    ASSERT(tag != NULL, "NULL tag in option_create_reqlang.");

    reqlang_opt->code = MC_REQLANG;
    reqlang_opt->len = len;
    memcpy(reqlang_opt->tag, tag, len);

    return reqlang_opt;
}

/**
 * Used to create a Multicast Scope List option.
 * @param len The total length of the data of the option when encoded.
 * @param scopes The list of scopes to store in the option.
 */
mc_option_scopelist *option_create_scopelist(uint16 len, list *scopes){
    mc_option_scopelist *scopelist_opt =
        (mc_option_scopelist *)malloc(sizeof(mc_option_scopelist));
    ASSERT(scopelist_opt != NULL, "Memory allocation failed.");
    ASSERT(scopes != NULL, "NULL list of scopes in option_create_scopelist.");

    scopelist_opt->code   = MC_SCOPELIST;
    scopelist_opt->len    = len;
    scopelist_opt->scopes = scopes;
    
    return scopelist_opt;
}

/**
 * Used to create a List of Address Ranges option.
 *
 * @param len The number of address ranges.
 * @param ranges The list of address ranges.
 * @return A List of Address Ranges option.
 */
mc_option_addrlist *option_create_addrlist(uint16 len, list *ranges){
    mc_option_addrlist *addrlist_opt =
        (mc_option_addrlist *)malloc(sizeof(mc_option_addrlist));
    ASSERT(addrlist_opt != NULL, "Memory allocation failed.");
    ASSERT(ranges != NULL, "NULL list of ranges in option_create_addrlist.");

    addrlist_opt->code = MC_ADDRLIST;
    addrlist_opt->len = len * 6;
    addrlist_opt->ranges = ranges;
    
    return addrlist_opt;
}

/**
 * Used to create a Current Time option.
 *
 * @param ctime The current time to use for the option. If it is 0, then the
 * current time as seen by the local clock will be used.
 * @return A Current Time option.
 */
mc_option_ctime *option_create_ctime(uint32 ctime){
    mc_option_ctime *ctime_opt =
        (mc_option_ctime *)malloc(sizeof(mc_option_ctime));
    ASSERT(ctime_opt != NULL, "Memory allocation failed.");

    ctime_opt->code = MC_CTIME;
    ctime_opt->len  = 4;
    if(ctime != 0) {
        ctime_opt->time = ctime;
    } else {
        ctime_opt->time = time(NULL);
    }
    return ctime_opt;
}

/**
 * Used to create a Feature List option.
 *
 * @param supported The list of supported options.
 * @param requested The list of requested options.
 * @param required The list of required options.
 * @return A Feature List option.
 */
extern mc_option_featurelist *option_create_featurelist(list *supported,
                                                        list *requested,
                                                        list *required){
    mc_option_featurelist *featurelist_opt =
        (mc_option_featurelist *)malloc(sizeof(mc_option_featurelist));
    ASSERT(featurelist_opt != NULL, "Memory allocation failed.");

    featurelist_opt->code = MC_FEATURELIST;
    featurelist_opt->supported = supported;
    featurelist_opt->requested = requested;
    featurelist_opt->required = required;

    return featurelist_opt;
}

/**
 * Used to create a Retry Time option.
 *
 * @param retry_time The retry time.
 * @return A Retry Time option.
 */
mc_option_rtime *option_create_rtime(uint32 retry_time){
    mc_option_rtime *rtime_opt =
        (mc_option_rtime *)malloc(sizeof(mc_option_rtime));
    ASSERT(rtime_opt != NULL, "Memory allocation failed.");

    rtime_opt->code = MC_RTIME;
    rtime_opt->len  = 4;
    rtime_opt->retry_time = retry_time;

    return rtime_opt;
}

/**
 * Used to create a Minimum Lease Time option.
 *
 * @param min_ltime The minimum lease time.
 * @return A Minimum Lease Time option.
 */
mc_option_min_ltime *option_create_min_ltime(uint32 min_ltime){
    mc_option_min_ltime *min_ltime_opt =
        (mc_option_min_ltime *)malloc(sizeof(mc_option_min_ltime));
    ASSERT(min_ltime_opt != NULL, "Memory allocation failed.");

    min_ltime_opt->code      = MC_MIN_LTIME;
    min_ltime_opt->len       = 4;
    min_ltime_opt->min_ltime = min_ltime;

    return min_ltime_opt;
}

/**
 * Used to create a Maximum Start Time option.
 *
 * @param max_stime The maximum start time.
 * @return A Maximum Start Time option.
 */
mc_option_max_stime *option_create_max_stime(uint32 max_stime){
    mc_option_max_stime *max_stime_opt =
        (mc_option_max_stime *)malloc(sizeof(mc_option_max_stime));
    ASSERT(max_stime_opt != NULL, "Memory allocation failed.");

    max_stime_opt->code      = MC_MAX_STIME;
    max_stime_opt->len       = 4;
    max_stime_opt->max_stime = max_stime;

    return max_stime_opt;
}

/**
 * Used to create an Error option.
 *
 * @param ecode The error code.
 * @param data The data of the error.
 * @param data_len The number of bytes of data.
 * @return An Error option.
 */
extern mc_option_error *option_create_error(uint16 ecode,
                                            uint8 *data,
                                            unsigned int data_len){
    mc_option_error *error_opt =
        (mc_option_error *)malloc(sizeof(mc_option_error));
    ASSERT(error_opt != NULL, "Memory allocation failed.");
    error_opt->code  = MC_ERROR;
    error_opt->len   = 2 + data_len;
    error_opt->ecode = ecode;
    error_opt->data  = (uint8 *)malloc(data_len * sizeof(uint8));
    ASSERT(error_opt->data != NULL, "Memory allocation failed.");
    memcpy(error_opt->data, data, data_len);

    return error_opt;
}

/**
 * This function is used to delete an option from memory.
 *
 * @param option The option to clear from memory.
 */
void option_delete(mc_option *option){
    ASSERT(option != NULL, "option is NULL in option_delete.");
    switch(option->code) {
    case MC_END:
        free((mc_option_end *)option);
        break;
    case MC_LTIME:
        free((mc_option_ltime *)option);
        break;
    case MC_SID:
        free((mc_option_sid *)option);
        break;
    case MC_LID:
        free(((mc_option_lid *)option)->id);
        free((mc_option_lid *)option);
        break;
    case MC_SCOPE:
        free((mc_option_scope *)option);
        break;
    case MC_REQLIST:
        list_first(((mc_option_reqlist *)option)->options);
        {
            uint16 *req = (uint16 *)list_delete(((mc_option_reqlist *)option)
                                                ->options);
            while(req != NULL) {
                free(req);
                req = (uint16 *)list_delete(((mc_option_reqlist *)option)
                                            ->options);
            }
        }
        list_destroy(((mc_option_reqlist *)option)->options);
        free((mc_option_reqlist *)option);
        break;
    case MC_STIME:
        free((mc_option_stime *)option);
        break;
    case MC_NUMADDR:
        free((mc_option_numaddr *)option);
        break;
    case MC_REQLANG:
        free(((mc_option_reqlang *)option)->tag);
        free((mc_option_reqlang *)option);
        break;
    case MC_SCOPELIST:
        if(((mc_option_scopelist *)option)->scopes != NULL) {
            scope *tmp_scope =
                (scope *)list_first(((mc_option_scopelist *)option)->scopes);
            while(tmp_scope != NULL) {
                if(tmp_scope->name_set != NULL) {
                    free(tmp_scope->name_set);
                }
                free(tmp_scope);
                tmp_scope =
                    (scope *)list_next(((mc_option_scopelist *)option)->scopes);
            }
            list_destroy(((mc_option_scopelist *)option)->scopes);
        }
        free((mc_option_scopelist *)option);
        break;
    case MC_ADDRLIST:
        list_first(((mc_option_addrlist *)option)->ranges);
        {
            address_range *range =
                (address_range *)list_delete(((mc_option_addrlist *)option)
                                             ->ranges);
            while(range != NULL) {
                free(range);
                range =
                    (address_range *)list_delete(((mc_option_addrlist *)option)
                                                 ->ranges);
            }
        }
        list_destroy(((mc_option_addrlist *)option)->ranges);
        free((mc_option_addrlist *)option);
        break;
    case MC_CTIME:
        free((mc_option_ctime *)option);
        break;
    case MC_FEATURELIST:
        if(((mc_option_featurelist *)option)->supported != NULL) {
            uint16 *tmp_code = (uint16 *)
                list_first(((mc_option_featurelist *)option)->supported);
            while(tmp_code != NULL) {
                free(tmp_code);
                tmp_code = (uint16 *)
                list_first(((mc_option_featurelist *)option)->supported);
            }
            list_destroy(((mc_option_featurelist *)option)->supported);
        }
        if(((mc_option_featurelist *)option)->requested != NULL) {
            uint16 *tmp_code = (uint16 *)
                list_first(((mc_option_featurelist *)option)->requested);
            while(tmp_code != NULL) {
                free(tmp_code);
                tmp_code = (uint16 *)
                list_first(((mc_option_featurelist *)option)->requested);
            }
            list_destroy(((mc_option_featurelist *)option)->requested);
        }
        if(((mc_option_featurelist *)option)->required != NULL) {
            uint16 *tmp_code = (uint16 *)
                list_first(((mc_option_featurelist *)option)->required);
            while(tmp_code != NULL) {
                free(tmp_code);
                tmp_code = (uint16 *)
                list_first(((mc_option_featurelist *)option)->required);
            }
            list_destroy(((mc_option_featurelist *)option)->required);
        }
        free((mc_option_featurelist *)option);
        break;
    case MC_RTIME:
        free((mc_option_rtime *)option);
        break;
    case MC_MIN_LTIME:
        free((mc_option_min_ltime *)option);
        break;
    case MC_MAX_STIME:
        free((mc_option_max_stime *)option);
        break;
    case MC_ERROR:
        free((mc_option_error *)option);
        break;
    default:
        ABORT("Unknown option in option_delete.");
        break;
    }
}
