/**
   The dll module represents a double linked list.
   @author Hkan Bystrm, Linus Svensson, ke stmark
   @version 2000-05-15
 */

#ifndef _DLL_H_
#define _DLL_H_

/* struct DLL is an opaque struct defined in dll.c */
typedef struct DLL DLL;

/* ELEMENT defined as a void pointer */
typedef void * ELEMENT;

/**
   Creates an empty DLL.
   @return a pointer to an ADT on success, NULL on failure
*/
DLL *createDLL();

/**
   Removes the ADT from memory.
   @return void.
*/
void destroyDLL(DLL *);

/**
   Checks to see if the list is empty.
   @return 1 if list is empty, 0 otherwise
*/
int isEmptyDLL(DLL *);

/**
   Sets "current" element to the next/prev/first/last element.
   @return 0 on success, -1 if such an element does not exist.
*/
int nextElement(DLL *);

/**
   Sets "current" element to the prev element.
   @return 0 on success, -1 if such an element does not exist.
*/
int prevElement(DLL *);

/**
   Sets "current" element to the first element.
   @return 0 on success, -1 if such an element does not exist.
*/
int firstElement(DLL *);

/**
   Sets "current" element to the last element.
   @return 0 on success, -1 if such an element does not exist.
*/
int lastElement(DLL *);

/**
   Gets the current element.
   @return "current" element, or NULL if such an element does not exist
*/
ELEMENT thisElement(DLL *);

/**
   Insert an element in the DLL.
   @precondition Inserts an element in the list. If where == 0 the element is inserted after "current" element, otherwise it is inserted before the current element.
   @postcondition The current element must be set to the inserted element.
   @return 0 on success, -1 on failure.
*/
int insertElement(DLL *, ELEMENT, int where);

/**
   Deletes "current" element.
   @postcondition New "current" element is set the following element if it exists. If not, it is set to the previous.
   @return 0 on succes, -1 on failure
*/
int deleteElement(DLL *);

/**
   Applies the function "f" on all elements in the list.
   @postcondition "current" element is not changed.
   @return all return values from the function calls bitwise OR'ed (|) together.
*/
int mapDLL(DLL *, int (*f)(ELEMENT));

/**
   Searches the list and for ELEMENT using the equal function.
   The search terminates when the equal-function returns a non-zero value,
   indicating that the ELEMENT was found in the list.
   @postcondition "current" element is set to the found element. If not found then "current" element is not changed.
   @return 0 if an element was found, otherwise -1.
*/
int searchDLL(DLL *, ELEMENT, int (*equal)(ELEMENT, ELEMENT));

/**
   The list is sorted in descending order by comparing elements in the list with the cmp-function. The cmp-function must return < 0 if the first element is smaller than the second, == 0 if the elements are equal and > 0 otherwise.
   @postcondition "current" element is not changed.
   @return void.
*/
void sortDLL(DLL *, int (*cmp)(ELEMENT, ELEMENT));

#endif

